
cdef extern from "liburing.h":

    cdef struct io_uring_sqe:
        pass

    cdef struct io_uring_cqe:
        unsigned long user_data
        int res
        unsigned int flags

    cdef struct io_uring_sq:
        pass
    
    cdef struct io_uring_cq:
        pass
    
    cdef struct io_uring:
        pass

    cdef struct io_uring_probe_op:
        unsigned char op
        unsigned short resv
        unsigned short flags
        unsigned int rev2

    cdef struct io_uring_probe:
        unsigned char last_op
        unsigned char ops_len
        unsigned short resv
        unsigned int resv2[3]
        io_uring_probe_op ops[255]

    cdef struct __kernel_timespec:
        pass

    cdef enum io_uring_op:
        IORING_OP_NOP
        IORING_OP_READV
        IORING_OP_WRITEV
        IORING_OP_FSYNC
        IORING_OP_READ_FIXED
        IORING_OP_WRITE_FIXED
        IORING_OP_POLL_ADD
        IORING_OP_POLL_REMOVE
        IORING_OP_SYNC_FILE_RANGE
        IORING_OP_SENDMSG
        IORING_OP_RECVMSG
        IORING_OP_TIMEOUT
        IORING_OP_TIMEOUT_REMOVE
        IORING_OP_ACCEPT
        IORING_OP_ASYNC_CANCEL
        IORING_OP_LINK_TIMEOUT
        IORING_OP_CONNECT
        IORING_OP_FALLOCATE
        IORING_OP_OPENAT
        IORING_OP_CLOSE
        IORING_OP_RSRC_UPDATE
        IORING_OP_STATX
        IORING_OP_READ
        IORING_OP_WRITE
        IORING_OP_FADVISE
        IORING_OP_MADVISE
        IORING_OP_SEND
        IORING_OP_RECV
        IORING_OP_OPENAT2
        IORING_OP_EPOLL_CTL
        IORING_OP_SPLICE
        IORING_OP_PROVIDE_BUFFERS
        IORING_OP_REMOVE_BUFFERS
        IORING_OP_TEE
        IORING_OP_SHUTDOWN
        IORING_OP_RENAMEAT
        IORING_OP_UNLINKAT
        IORING_OP_MKDIRAT
        IORING_OP_SYMLINKAT
        IORING_OP_LINKAT
        IORING_OP_MSG_RING
        IORING_OP_FSETXATTR
        IORING_OP_SETXATTR
        IORING_OP_FGETXATTR
        IORING_OP_GETXATTR
        IORING_OP_SOCKET
        IORING_OP_URING_CMD
        IORING_OP_SENDZC_NOTIF
        IORING_OP_LAST
    
    cdef io_uring_probe* io_uring_get_probe_ring(io_uring*);
    cdef io_uring_probe* io_uring_get_probe();
    cdef int io_uring_opcode_supported(io_uring*, io_uring_op);

    cdef void io_uring_free_probe(io_uring_probe*);
    cdef int io_uring_opcode_supported(const io_uring_probe*, int);

    cdef int io_uring_queue_init(unsigned int entries, io_uring *ring, unsigned int flags);
    cdef int io_uring_queue_exit(io_uring *ring);
    
    cdef int io_uring_wait_cqes(io_uring *ring, io_uring_cqe **cqe_ptr, unsigned int wait_nr, __kernel_timespec *ts);
    cdef int io_uring_submit(io_uring*);

    cdef void io_uring_cq_advance(io_uring *ring, unsigned int nr);
    cdef void io_uring_cqe_seen(io_uring *ring, io_uring_cqe *cqe);

    cdef void io_uring_sqe_set_data(io_uring_sqe*, void*);
    cdef void *io_uring_cqe_get_data(const io_uring_cqe *cqe);

    cdef unsigned int io_uring_sq_ready(const io_uring *ring);
    cdef unsigned int io_uring_sq_space_left(const io_uring *ring);
    cdef unsigned int io_uring_cq_ready(const io_uring *ring);
    cdef int io_uring_peek_cqe(io_uring *ring, io_uring_cqe **cqe_ptr);
    cdef io_uring_sqe* io_uring_get_sqe(io_uring *ring);

    cdef int io_uring_register_eventfd(io_uring *ring, int fd);

    # NOP
    cdef void io_uring_prep_nop(
        io_uring_sqe *sqe
    );
    #FILE IO AND FILESYSTEM
    cdef void io_uring_prep_read(
        io_uring_sqe *sqe,
        int fd,
        void *buf,
        unsigned int nbytes,
        unsigned long offset
    );
    cdef void io_uring_prep_write(
        io_uring_sqe *sqe,
        int fd,
        void *buf,
        unsigned int nbytes,
        unsigned long offset
    );
    cdef void io_uring_prep_statx(
        io_uring_sqe *sqe,
        int dfd,
        const char *buf,
        int flags,
        unsigned int mask,
        void* statxbuf # actually a pointer to a statx struct
    );
    cdef void io_uring_prep_openat(
        io_uring_sqe *sqe,
        int dfd,
        const char *pathname,
        int flags,
        int mode  
    );
    cdef void io_uring_prep_close(
        io_uring_sqe *sqe,
        int fd
    );
    cdef void io_uring_prep_renameat(
        io_uring_sqe *sqe,
        int olddfd,
        const char *oldpath,
        int newdfd,
        const char *newpath,
        int flags
    );
    cdef void io_uring_prep_unlinkat(
        io_uring_sqe *sqe,
        int dfd,
        const char *pathname,
        int flags
    );

    # NETWORK IO AND SOCKETS
    cdef void io_uring_prep_accept(
        io_uring_sqe *sqe,
        int sockfd,
        void *addr, # struct sockaddr *
        unsigned int *addrlen,
        int flags
    );
    cdef void io_uring_prep_connect(
        io_uring_sqe *sqe,
        int sockfd,
        void *addr,
        unsigned int addrlen # actually socklen_t
    );
    cdef void io_uring_prep_recv(
        io_uring_sqe *sqe,
        int sockfd,
        void *buf,
        unsigned int len,
        int flags
    );
    cdef void io_uring_prep_send(
        io_uring_sqe *sqe,
        int sockfd,
        void *buf,
        unsigned int len,
        int flags
    );
    cdef void io_uring_prep_socket(
        io_uring_sqe *sqe,
        int domain,
        int type_,
        int protocol,
        unsigned int flags
    );
    # SPLICE
    cdef void io_uring_prep_splice(
        io_uring_sqe *sqe,
        int fd_in,
        long off_in,
        int fd_out,
        long off_out,
        unsigned int len,
        unsigned int flags
    );

    # REQUEST MANAGEMENT
    cdef void io_uring_prep_cancel(
        io_uring_sqe *sqe,
        void* user_data,
        int flags
    );